#ifndef POTENTIEL_H
#define POTENTIEL_H

#include "matr9f.h"
#include "octree.h"
#include "patrons.h"
#include <math.h>

#define PRECISION_DERIVEE       0x1p-12 //2^-12
#define PRECISION_DERIVEE_1     0x1p+12 //2^+12

class Pot0{
    public:
    virtual float operator()(Vect3f const&a){return 0;}
    Vect3f getgrad(Vect3f pos){
        float p0 = (*this)(pos);
        return PRECISION_DERIVEE_1*Vect3f((*this)(pos+Vect3f(PRECISION_DERIVEE,0,0))-p0,
                                        (*this)(pos+Vect3f(0,PRECISION_DERIVEE,0))-p0,
                                        (*this)(pos+Vect3f(0,0,PRECISION_DERIVEE))-p0);
    }

    /* méthode de Newton pour plaquer le vecteur sur la surface:
    La méthode peut diverger si le point est initialement trop loin de la surface,
    dans ce cas là il peut apparaitre un trou dans la surface, c'est que les composantes du vecteur sont à "nan"  */
    void plaquevect(Vect3f &pos, float ref);
};


/* un exemple de potentiel. Si vous voulez créer votre propre potentiel,
utilisez une classe héritée de Pot0 avec une fonction virtuelle du type:
float operator()(Vect3f const&a); */
class Verre : public Pot0{
    public:
    virtual float operator()(Vect3f const&a){return -(a.x*a.x+a.z*a.z-carre(log(a.y+3.2))-0.02);}
};


typedef Octree<Indicestab> OctreeIndices;

// le potentiel du blob a une forme d'ellipsoïde, gouvernée par une matrice 3*3
class Blob{
    public:
    inline Blob():pos(),influence(){}// constructeur par défault: blob infini (car matrice influence nulle)
    inline Blob(Vect3f const&centre,Matr9f const&forme){pos=centre;influence=forme;}

    //la fonction qui donne le potentiel pour le blob
    float operator()(Vect3f const&a)const{
        Vect3f v = influence*(a-pos);
        return carre(max(1.-v*v,0.));// raccord C1 vers la fonction nulle
        // le blob a une portée limitée, plus les coefficients de la matrice sont élevée, plus le blob est petit
    }
    Vect3f pos;// position du blob
    Matr9f influence;// forme du blob

    // permet de savoir quelle portée maximale va avoir le blob.
    inline float getlargeurzone()const{return (!influence).norme();}
};
typedef dynamictab<Blob> Blobtab;

class Potentiel : public Pot0{
    public:
    OctreeIndices oct;// les indices des blobs sont rangés dans un octree pour plus d'efficacité
    Blobtab tab;// le tableau de blob sur lequel les éléments de l'octree font référence

    Traceur<Indicestab> trac;// le navigateur à octree

    // constructeurs à partir de la largeur de l'octree et de l'offset
    inline Potentiel(Vect3f const&t, float echelle):oct(),tab(),trac(oct,echelle,t){}
    inline Potentiel(float largeur):oct(),tab(),trac(oct,largeur/POSMAX,largeur*Vect3f(.5,.5,.5)){}

    // permet d'appliquer une transformation linéaire aux blobs
    Potentiel(Potentiel const&p, Matr9f const&transformation);

    // constructeur copie
    Potentiel(Potentiel const&p):oct(p.oct),tab(p.tab),trac(oct,p.trac.getscale(),p.trac.getoffset()){}

    // constructeur à partir de la largeur et de l'offset d'un traceur
    template<class T>
    inline Potentiel(Traceur<T> const&t):oct(),tab(),trac(oct,t.getscale(),t.getoffset()){}

    void addblob(Blob const&b);//ajoute un blob au champ de blobs


    float operator()(Vect3f const&a);// fonction potentielle
};

// convertion des couleurs sous forme d'entier (int) à vecteur de 3 flottants (Vect3f)
inline Vect3f couleur2Vect3f(int c){return Vect3f((c>>16)&0xff,(c>>8)&0xff,c&0xff);}
inline int Vect3f2couleur(Vect3f c){return 0xff000000 + ((int)c.x<<16)+((int)c.y<<8)+(int)c.z;}

// Classe champ de couleur
class Couleurs{
    public:

    class Zone{
        public:
        inline Zone():b(),a(){}
        inline Zone(Blob const&b0, Vect3f const&a0):b(b0),a(a0){}
        Blob b;
        Vect3f a;
    };
    OctreeIndices oct;
    dynamictab<Zone> tab;

    Traceur<Indicestab> trac;

    inline Couleurs(Vect3f const&t, float echelle):oct(),tab(),trac(oct,echelle,t){}
    template<class T2>
    inline Couleurs(Traceur<T2> const&t):oct(),tab(),trac(oct,t.getscale(),t.getoffset()){}

    void addcouleur(Blob const&b, int couleur);

    void addinfinitecouleur(int couleur){// colore
        tab.add(Zone(Blob(),couleur2Vect3f(couleur)));// ajout du blob dans le tableau
        oct.node.add(tab.end()-1);// ajout du blob à la racine de l'octree
    }
    int operator()(Vect3f const&a);// retourne la couleur dans un entier
};


#endif
